/*
  Copyright 2012 Marvin Pinto (me@marvinp.ca)

  Licensed under the Apache License, Version 2.0 (the "License"); you may not
  use this file except in compliance with the License.  You may obtain a copy of
  the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
  License for the specific language governing permissions and limitations under
  the License.

  This product includes the Apache Commons Lang 3.1 library developed by the
  Apache Software Foundation, under the Apache License 2.0.

  This product includes the Gson 2.1 library developed by Google Inc., under the
  Apache License 2.0.

  This product includes the JSoup 1.6.1 library developed by Jonathan Hedley,
  under the MIT license.
*/
package ca.marvinp.se;

import ca.marvinp.se.entities.Question;
import ca.marvinp.se.entities.WeightedQuestion;
import ca.marvinp.se.utils.HttpQuery;
import ca.marvinp.se.utils.Date;
import ca.marvinp.se.utils.ApiTracker;
import ca.marvinp.se.utils.ProgressBar;
import ca.marvinp.se.misc.Settings;
import ca.marvinp.se.misc.Constants;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import com.google.gson.Gson;
import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.text.WordUtils;
import org.jsoup.Jsoup;
import java.text.DecimalFormat;
import java.text.NumberFormat;



public class InterestingQuestions {

  private Settings settings;


  public InterestingQuestions(Settings s) {
    this.settings = s;
    // settings = new Settings();

    // Gather the list of weighted questions
    List<WeightedQuestion> qs = gatherInterestingQuestions();

    // We encountered an error generating the list, no point in continuing
    if (qs == null) {
      return;
    }

    // Sort the list of questions
    Collections.sort(qs);

    // Finally print out the list of the lowest weighted questions
    System.out.println(settings.getLineSeparator()
                       + settings.getLineSeparator());
    printInterestingQuestions(qs);

    // Print out the total number of API calls made
    ApiTracker.printSummary();
  }







  private void printInterestingQuestions(List<WeightedQuestion> questions) {
    NumberFormat formatter = new DecimalFormat("#0.000");
    for (int i=0;i<questions.size();i++) {

      // Break out of the loop when we're done printing user specified number of
      // questions
      if (i >= settings.getNumDisplayQuestions()) {
        break;
      }

      WeightedQuestion wq = questions.get(i);
      StringBuilder sb = new StringBuilder();

      // Print question header
      sb.append(getDashes() + settings.getLineSeparator() );
      sb.append("Q - " + cleanString( wq.getQuestion().getTitle(),
                                      Constants.WRAP_COLUMN - "Q - ".length() )
                + settings.getLineSeparator() );

      sb.append(getDashes() + settings.getLineSeparator() );

      // Add this question to the "seen" list
      settings.addToSeenList(wq.getQuestion().getQuestionId());

      // Misc information about the question
      sb.append("URL: " +settings.getSiteUrl()+ "/q/"
                + wq.getQuestion().getQuestionId()
                + settings.getLineSeparator() );
      sb.append("Date: " + Date.getFormattedDate
                (wq.getQuestion().getCreationDate()) + ", ");
      sb.append("Votes: "
                + wq.getQuestion().getScore() + " ("
                + "+" + wq.getQuestion().getUpVoteCount() + "/"
                + "-" + wq.getQuestion().getDownVoteCount() + "), ");
      sb.append("MScore: " + formatter.format(wq.getWeight())
                + settings.getLineSeparator() );
      sb.append("Author: " + wq.getQuestion().getOwner().getDisplayName()
                + " (" + wq.getQuestion().getOwner().getAcceptRate()
                + "% accept rate), Rep: "
                + wq.getQuestion().getOwner().getReputation()
                + settings.getLineSeparator() );
      sb.append("Answers: " + wq.getTotalAnswers() + ", "
                + "Comments: " + wq.getTotalComments() + ", "
                + "Views: " + wq.getQuestion().getViewCount() + ", "
                + "Favorites: " + wq.getQuestion().getFavoriteCount()
                + settings.getLineSeparator() );


      // Question tags
      String tags = prettifyTags( wq.getQuestion().getTags() );
      sb.append("Tags: ");
      sb.append(WordUtils.wrap(tags, Constants.WRAP_COLUMN,
                               settings.getLineSeparator(), true )
                + settings.getLineSeparator() );


      // Question preview
      sb.append("Preview: ----" + settings.getLineSeparator() );
      String b = cleanString(wq.getQuestion().getBody(), Constants.WRAP_COLUMN);
      String body;
      if (b.length() > Constants.PREVIEW_LIMIT) {
        body = b.substring(0, Constants.PREVIEW_LIMIT);
        body = body + " (...)";
      } else {
        body = b.substring(0, b.length());
      }
      sb.append(cleanString(body, Constants.WRAP_COLUMN)
                + settings.getLineSeparator());



      // Finally print out the whole question
      System.out.println(sb.toString());
    }
  }




  private static String getDashes() {
    StringBuilder sb = new StringBuilder();
    for (int i=0;i<Constants.WRAP_COLUMN;i++) {
      sb.append("-");
    }
    return sb.toString();
  }






  private String cleanString(String input, int wrapColumn) {
    input = StringEscapeUtils.unescapeHtml4(input);
    input = Jsoup.parse(input).text();
    input = input.replaceAll("\\s+", " ");
    input = WordUtils.wrap(input, wrapColumn,
                           settings.getLineSeparator(), true );
    return input;
  }








  private static String prettifyTags(List<String> tags) {
    StringBuilder sb = new StringBuilder();
    if (tags == null) { return ""; }
    for (int i=0;i<tags.size();i++) {
      if (i == (tags.size()-1)) {
        // We're on the last tag
        sb.append("[" + tags.get(i) + "]");
      } else {
        sb.append("[" + tags.get(i) + "] ");
      }
    }
    return sb.toString();
  }








  private boolean vetQuestion(Question q) {

    // Check to see if this question is in the "seen" list
    if ( settings.isInSeenList(q.getQuestionId()) ) {
      return false;
    }

    // Already has an accepted answer
    if ( settings.getRestrAcceptedAnswer() ) {
      if (q.getAcceptedAnswerId() != 0) {
        return false;
      }
    }

    // We only care about questions from registered users and moderators
    if ( settings.getRestrUser() ) {
      if ( ( ! q.getOwner().getUserType().equals("registered") ) 
           && ( ! q.getOwner().getUserType().equals("moderator") ) ) {
        return false;
      }
    }

    // We don't care about closed questions
    if ( settings.getRestrClosedQuestion() ) {
      if ( q.getClosedDate() != -1 ) {
        return false;
      }
    }

    // Ignore the question if it only has one tag (indicative of not well
    // thought out questions.
    List<String> qTags = q.getTags();
    if ( settings.getRestrSingleTag() ) {
      if (qTags == null) {
        // No tags whatsoever, why bother. Should note that we'll probably never
        // reach this case, but it doesn't hurt
        return false;
      } else if (qTags.size() == 1) {
        return false;
      }
    }

    // Ignore questions in our banned tag list
    List<String> bTags = settings.getIgnoredTags();
    if (bTags != null) {
      for (int i=0;i<bTags.size();i++) {
        // Does a substring match
        for (int x=0;x<qTags.size();x++) {
          String qTag = qTags.get(x);
          String bTag = bTags.get(i);
          if ( qTag.indexOf(bTag) >= 0 ) {
            return false;
          }
        }
      }
    }

    return true;
  }







  private List<WeightedQuestion> gatherInterestingQuestions() {
    ProgressBar.update(Constants.PRGB_REG);
    int page = 1;
    boolean moreResults = true;
    int questionCount = 0;
    List<WeightedQuestion> weightedQuestions =
      new ArrayList<WeightedQuestion>();


    while (moreResults) {
      String url;
      ProgressBar.update(Constants.PRGB_REG);

      if (settings.getApiKey() == null) {
        url = "https://api.stackexchange.com/2.0/search?"
          + "&page=" + page
          + "&pagesize=" + settings.getQueryPagesize()
          + "&fromdate=" + settings.getFromDate()
          + "&todate=" + settings.getToDate()
          + "&order=" + settings.getOrder()
          + "&sort=" + Constants.QUERY_SORT_FIELD
          + "&tagged=" +settings.getTags()
          + "&site=" + settings.getSite()
          + "&filter=" + Constants.Q_FILTER;
      } else {
        url = "https://api.stackexchange.com/2.0/search?"
          + "key=" + settings.getApiKey()
          + "&page=" + page
          + "&pagesize=" + settings.getQueryPagesize()
          + "&fromdate=" + settings.getFromDate()
          + "&todate=" + settings.getToDate()
          + "&order=" + settings.getOrder()
          + "&sort=" + Constants.QUERY_SORT_FIELD
          + "&tagged=" +settings.getTags()
          + "&site=" + settings.getSite()
          + "&filter=" + Constants.Q_FILTER;
      }

      String queryData = HttpQuery.getData(url);

      if (queryData == null) {
        System.out.println();
        System.out.println(Constants.ERR_NO_DATA);
        System.out.println(Constants.ERR_NO_DATA2);
        return null;
      }

      IQResults res = new Gson().fromJson(queryData, IQResults.class);
      List<Question> questions = res.getItems();
      ApiTracker.setMaxQuota(res.getQuotaMax());
      ApiTracker.setQuotaRemaining(res.getQuotaRemaining());

      for (int i=0;i<questions.size();i++) {
        Question q = questions.get(i);

        // Move on if we don't like this question
        if ( !vetQuestion(q) ) {
          continue;
        }

        weightedQuestions.add(new WeightedQuestion(q, settings));

        // Break out of the inner loop if we've reached (or passed) the maximum
        // number of questions we're allowed to process
        if (questionCount >= settings.getMaxQuestions()) {
          break;
        }
        questionCount++;
      }

      // Break out of the main loop if we've reached (or passed) the maximum
      // number of questions we're allowed to process
      if (questionCount >= settings.getMaxQuestions()) {
        break;
      }
      moreResults = res.hasMore();
      page++;
    }

    return weightedQuestions;
  }






  private class IQResults {
    private List<Question> items;
    private boolean has_more;
    private int quota_max;
    private int quota_remaining;
    public List<Question> getItems() { return items; }
    public boolean hasMore() { return has_more; }
    public int getQuotaMax() { return quota_max; }
    public int getQuotaRemaining() { return quota_remaining; }
  }


}




