/*
  Copyright 2012 Marvin Pinto (me@marvinp.ca)

  Licensed under the Apache License, Version 2.0 (the "License"); you may not
  use this file except in compliance with the License.  You may obtain a copy of
  the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
  License for the specific language governing permissions and limitations under
  the License.
*/
package ca.marvinp.se.entities;

import ca.marvinp.se.entities.Question;
import ca.marvinp.se.entities.ShallowUser;
import ca.marvinp.se.entities.Answer;
import ca.marvinp.se.utils.Date;
import ca.marvinp.se.misc.Constants;
import ca.marvinp.se.misc.Settings;
import java.util.List;



public class WeightedQuestion implements Comparable<WeightedQuestion> {

  private float weight = 0;
  private Question question;
  private int totalAnswers = 0;
  private int totalComments = 0;
  private Settings settings;


  public WeightedQuestion(Question q, Settings s) {
    this.question = q;
    this.settings = s;
    calculateWeight();
  }


  // Getters
  public float getWeight() { return weight; }
  public Question getQuestion() { return question; }
  public int getTotalAnswers() { return totalAnswers; }
  public int getTotalComments() { return totalComments; }





  private void calculateWeight() {
    weight = weight + weighQuestion();
    weight = weight + weighAnswers();
    weight = weight / Constants.WEIGHT_SCALE;
  }





  private float weighAnswers() {
    float mWeight = 0;

    List<Answer> answers = question.getAnswers();
    if (answers == null) {
      // In the case that there are no answers (yet)
      return mWeight;
    }

    totalAnswers = answers.size();

    for (int i=0;i<totalAnswers;i++) {

      // Essentially, the more answers this question receives, the more heavily
      // weighted the 2nd, 3rd, etc answer will be.

      // factor in the answerer's rep
      int rep = answers.get(i).getOwner().getReputation();
      mWeight = mWeight + (float)(settings.getPenAnswerOwnerRep()
                                  * rep * (i+1) );

      // factor in the upvotes and downvotes
      float upvotes = answers.get(i).getUpVoteCount() *
        settings.getPenAnswerUpvotes() * (i+1);
      float downvotes = answers.get(i).getDownVoteCount() *
        settings.getPenAnswerDownvotes() * (i+1);
      mWeight = mWeight + upvotes;
      mWeight = mWeight + downvotes;

      // factor in the number of comments this answer has received
      List<Comment> comments = answers.get(i).getComments();
      if (comments != null) {
        int cs = 0;
        for (int x=0;x<comments.size();x++) {
          // Factor in the upvotes for each comment, as well as the comment
          // itself
          cs = cs + comments.get(x).getScore();
          cs++;
        }
        totalComments = totalComments + cs;
        mWeight = mWeight + (float)(settings.getPenAnswerComments()
                                    * cs * (i+1));
      }

    }

    return mWeight;
  }







  private float weighQuestion() {
    float mWeight = 0;

    // factor in the question owner's rep
    int rep = question.getOwner().getReputation();
    mWeight = mWeight + (float)(0.001 * settings.getPenQuestionOwnerRep());

    // factor in the upvotes and downvotes
    float upvotes = question.getUpVoteCount() *
      settings.getPenQuestionUpvotes();
    float downvotes = question.getDownVoteCount() *
      settings.getPenQuestionDownvotes();
    mWeight = mWeight + upvotes;
    mWeight = mWeight + downvotes;

    // factor in the question owner's accept rate
    int aRate = question.getOwner().getAcceptRate();
    mWeight = mWeight + (float)((100 - aRate) *
                                settings.getPenQuestionOwnerAcceptRate());

    // factor in the number of views this question has had
    int vc = question.getViewCount();
    mWeight = mWeight + (float)(settings.getPenQuestionViewcount() * vc);

    // factor in the number of times this question has been favorited
    int fav = question.getFavoriteCount();
    mWeight = mWeight + (float)(settings.getPenQuestionFavcount() * fav);

    // factor in the number of comments this question has received
    List<Comment> comments = question.getComments();
    if (comments != null) {
      int cs = 0;
      for (int x=0;x<comments.size();x++) {
        // Factor in the upvotes for each comment, as well as the comment
        // itself
        cs = cs + comments.get(x).getScore();
        cs++;
      }
      mWeight = mWeight + (float)(settings.getPenQuestionComments() * cs);
    }

    // factor in the length of the question
    String body = question.getBody();
    mWeight = mWeight + (float)(settings.getPenQuestionCharcount()
                                * body.length());

    // factor in the age of this question
    long currDate = Date.getCurrentDate();
    long qDate = question.getCreationDate();
    mWeight = mWeight + (float)((currDate - qDate) *
                                settings.getPenQuestionAge());

    // factor in the bounty prize (if any) for this question
    mWeight = mWeight + (float)(question.getBountyAmount() *
                                settings.getPenQuestionBounty());

    return mWeight;
  }







  public int compareTo(WeightedQuestion otherQuestion) {

    final int LESSTHAN = -1;
    final int EQUAL = 0;
    final int GREATERTHAN = 1;

    // Object comparison - doesn't hurt to optimize a bit
    if (this == otherQuestion) {
      return EQUAL;
    }

    float otherWeight = otherQuestion.getWeight();
    int compRes = Float.compare(weight, otherWeight);
    if (compRes > 0) {
      // My weight is greater
      return GREATERTHAN;
    } else if (compRes < 0) {
      // Other weight is greater
      return LESSTHAN;
    } else {
      // We're both equal
      return EQUAL;
    }

  }


}


