/*
  Copyright 2012 Marvin Pinto (me@marvinp.ca)

  Licensed under the Apache License, Version 2.0 (the "License"); you may not
  use this file except in compliance with the License.  You may obtain a copy of
  the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
  License for the specific language governing permissions and limitations under
  the License.

  This product includes the Gson 2.1 library developed by Google Inc., under the
  Apache License 2.0.
*/
package ca.marvinp.se.misc;

import ca.marvinp.se.utils.Date;
import ca.marvinp.se.misc.Constants;
import ca.marvinp.se.entities.Site;
import ca.marvinp.se.utils.HttpQuery;
import ca.marvinp.se.utils.ProgressBar;
import ca.marvinp.se.utils.ApiTracker;
import java.util.List;
import java.util.ArrayList;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import java.io.FileWriter;
import com.google.gson.reflect.TypeToken;
import java.lang.reflect.Type;



public class Settings {

  private int max_questions = 300;
  private int display_questions = 10;
  private String api_key = null;
  private String site = "stackoverflow";
  private String site_url = null;
  private List<String> wanted_tags;
  private List<String> ignored_tags;
  private List<Integer> seenQuestions;
  private String order = "desc";
  private int from_date;
  private int to_date;
  private long fromDateConv;
  private long toDateConv;
  private int query_pagesize = 50;
  private String already_seen_file = null;

  // Misc restrictions
  private boolean accepted_answer_restriction = true;
  private boolean user_restriction = true;
  private boolean closed_question_restriction = true;
  private boolean single_tag_restriction = true;

  // Penalties associated with questions
  private float q_owner_rep = 0.001f;
  private float q_upvote = -1.0f;
  private float q_downvote = 5.0f;
  private float q_owner_accept_rate = 2.0f;
  private float q_viewcount = -0.0005f;
  private float q_favcount = -0.01f;
  private float q_comments = 0.5f;
  private float q_charcount = 0.00f;
  private float q_age = -0.0001f;
  private float q_bounty = -0.10f;

  // Penalties associated with answers
  private float a_owner_rep = 0.01f;
  private float a_upvote = 5.0f;
  private float a_downvote = -1.0f;
  private float a_comments = 0.25f;





  public boolean areSettingsValid() {

    ProgressBar.update(Constants.PRGB_REG);
    // Make sure wanted_tags contains something
    if (wanted_tags == null) {
      return false;
    } else if (wanted_tags.size() < 1) {
      return false;
    }

    // Ensure the range dates are in order
    ProgressBar.update(Constants.PRGB_REG);
    if (from_date >= to_date) {
      return false;
    } else if (from_date > -1) {
      return false;
    } else if (to_date > 0) {
      return false;
    }
    fromDateConv = Date.getCurrentDate() - (Math.abs(from_date) *
                                            Constants.ONE_DAY);
    if (to_date == 0) {
      toDateConv = Date.getCurrentDate();
    } else {
      toDateConv = Date.getCurrentDate() - (Math.abs(to_date) *
                                            Constants.ONE_DAY);
    }

    // Ensure 0 < pagesize <= 100
    ProgressBar.update(Constants.PRGB_REG);
    if (query_pagesize < 1 || query_pagesize > 100) {
      return false;
    }

    // Ensure order is one of: desc, asc
    ProgressBar.update(Constants.PRGB_REG);
    if ( !order.equals("desc") && !order.equals("asc") ) {
      return false;
    }

    // Retrieve the site URL
    ProgressBar.update(Constants.PRGB_REG);
    if (!retrieveSiteUrl()) {
      System.out.println(Constants.ERR_API_COMM);
      return false;
    }

    // Load the seen questions from disk
    ProgressBar.update(Constants.PRGB_REG);
    if (!loadSeenQuestions()) {
      System.out.println(Constants.ERR_SEEN_READ);
      return false;
    }

    return true;
  }







  private boolean loadSeenQuestions() {

    ProgressBar.update(Constants.PRGB_REG);
    if (already_seen_file == null) {
      // Not using the "seen" file, we can ignore and return
      return true;
    }

    ProgressBar.update(Constants.PRGB_REG);
    boolean exists = (new File(already_seen_file)).exists();
    if (!exists) {
      // File doesn't exist (yet), initialize the seenQuestions list and move on
      seenQuestions = new ArrayList<Integer>();
      return true;
    }

    ProgressBar.update(Constants.PRGB_REG);
    try {
      ProgressBar.update(Constants.PRGB_REG);
      BufferedReader br = new BufferedReader(new FileReader(already_seen_file));
      Type type = new TypeToken<List<Integer>>(){}.getType();
      seenQuestions = new Gson().fromJson(br, type);
      ProgressBar.update(Constants.PRGB_REG);
    } catch (Exception e) {
      return false;
    }

    ProgressBar.update(Constants.PRGB_REG);
    return true;
  }







  public void writeSeenQuestionstoFile() {
    if (already_seen_file == null) {
      // Not using the "seen" file, we can ignore and return
      return;
    }

    try {
      // Gson gson = new GsonBuilder().setPrettyPrinting().create();
      // String json = gson.toJson(seenQuestions);

      Gson gson = new Gson();
      String json = gson.toJson(seenQuestions);

      FileWriter writer = new FileWriter(already_seen_file);
      writer.write(json);
      writer.close();
    } catch (Exception e) {
      System.out.println();
      System.out.println(Constants.ERR_SEEN_WRITE);
    }
  }





  public boolean isInSeenList(int q) {
    if (seenQuestions!=null) {
      return seenQuestions.contains(q);
    }
    return false;
  }





  public void addToSeenList(int q) {
    if (seenQuestions!=null) {
      seenQuestions.add(q);
    }
  }






  private boolean retrieveSiteUrl() {
    ProgressBar.update(Constants.PRGB_REG);
    String url = "";
    if (getApiKey() == null) {
      url = "https://api.stackexchange.com/2.0/sites?pagesize=100000"
        + "&filter=" + Constants.S_FILTER;
    } else {
      url = "https://api.stackexchange.com/2.0/sites?pagesize=100000"
        + "&key=" + getApiKey()
        + "&filter=" + Constants.S_FILTER;
    }

    ProgressBar.update(Constants.PRGB_REG);
    String queryData = HttpQuery.getData(url);

    ProgressBar.update(Constants.PRGB_REG);
    if (queryData == null) {
      System.out.println();
      System.out.println(Constants.ERR_NO_DATA);
      System.out.println(Constants.ERR_NO_DATA2);
      return false;
    }

    ProgressBar.update(Constants.PRGB_REG);
    SQResults res = new Gson().fromJson(queryData, SQResults.class);
    List<Site> sites = res.getItems();
    ApiTracker.setMaxQuota(res.getQuotaMax());
    ApiTracker.setQuotaRemaining(res.getQuotaRemaining());
    if (sites == null) {
      System.out.println();
      System.out.println(Constants.ERR_BAD_DATA);
      return false;
    }

    ProgressBar.update(Constants.PRGB_REG);
    for (int i=0;i<sites.size();i++) {
      ProgressBar.update(Constants.PRGB_REG);
      String ss = sites.get(i).getApiSiteParameter();
      if (ss.equals(getSite())) {
        site_url = sites.get(i).getSiteUrl();
        break;
      }
    }
    return true;
  }






  // Getters
  public String getApiKey() { return Constants.API_KEY; }
  public int getMaxQuestions() { return max_questions; }
  public int getNumDisplayQuestions() { return display_questions; }
  public String getSite() { return site; }
  public String getSiteUrl() { return site_url; }
  public String getOrder() { return order; }
  public long getFromDate() { return fromDateConv; }
  public long getToDate() { return toDateConv; }
  public String getLineSeparator() {
    return System.getProperty("line.separator");
  }
  public List<String> getIgnoredTags() { return ignored_tags; }


  public String getTags() {
    StringBuilder sb = new StringBuilder();
    for (int i=0;i<wanted_tags.size();i++) {
      if ( i == (wanted_tags.size()-1) ) {
        sb.append(wanted_tags.get(i));
      } else {
        sb.append(wanted_tags.get(i) + ";");
      }
    }
    return sb.toString();
  }


  public int getQueryPagesize() { return query_pagesize; }

  public float getPenQuestionOwnerRep() { return q_owner_rep; }
  public float getPenQuestionUpvotes() { return q_upvote; }
  public float getPenQuestionDownvotes() { return q_downvote; }
  public float getPenQuestionOwnerAcceptRate() { return q_owner_accept_rate; }
  public float getPenQuestionViewcount() { return q_viewcount; }
  public float getPenQuestionFavcount() { return q_favcount; }
  public float getPenQuestionComments() { return q_comments; }
  public float getPenQuestionCharcount() { return q_charcount; }
  public float getPenQuestionAge() { return q_age; }
  public float getPenQuestionBounty() { return q_bounty; }

  public float getPenAnswerOwnerRep() { return a_owner_rep; }
  public float getPenAnswerUpvotes() { return a_upvote; }
  public float getPenAnswerDownvotes() { return a_downvote; }
  public float getPenAnswerComments() { return a_comments; }


  public boolean getRestrAcceptedAnswer() {
    return accepted_answer_restriction;
  }
  public boolean getRestrUser() { return user_restriction; }
  public boolean getRestrClosedQuestion() {
    return closed_question_restriction;
  }
  public boolean getRestrSingleTag() {
    return single_tag_restriction;
  }



  private class SQResults {
    private List<Site> items;
    private boolean has_more;
    private int quota_max;
    private int quota_remaining;
    public List<Site> getItems() { return items; }
    public boolean hasMore() { return has_more; }
    public int getQuotaMax() { return quota_max; }
    public int getQuotaRemaining() { return quota_remaining; }
  }

}

